﻿using System;
using System.Collections.Generic;
using System.Reflection;
using HeuristicLab.Persistence.Interfaces;

namespace HeuristicLab.Persistence.Core {
  
  public class ConfigurationService {

    private static ConfigurationService instance;
    private Dictionary<IFormat, Configuration> customConfigurations;
    public readonly Dictionary<IFormat, List<IFormatter>> Formatters;
    public readonly List<IDecomposer> Decomposers;
    
    public static ConfigurationService Instance {
      get {
        if (instance == null)
          instance = new ConfigurationService();
        return instance;
      }
    }

    public ConfigurationService() {
      Formatters = new Dictionary<IFormat, List<IFormatter>>();
      Decomposers = new List<IDecomposer>();
      customConfigurations = new Dictionary<IFormat, Configuration>();
      Reset();
    }

    public void Reset() {      
      customConfigurations.Clear();
      Assembly defaultAssembly = Assembly.GetExecutingAssembly();
      DiscoverFrom(defaultAssembly);
      foreach (Assembly a in AppDomain.CurrentDomain.GetAssemblies())
        if ( a != defaultAssembly )
          DiscoverFrom(a);
    }

    protected void DiscoverFrom(Assembly a) {
      foreach (Type t in a.GetTypes()) {
        if (t.GetInterface(typeof (IFormatter).FullName) != null) {
          try {
            IFormatter formatter = (IFormatter) Activator.CreateInstance(t, true);
            if ( ! Formatters.ContainsKey(formatter.Format) ) {
              Formatters.Add(formatter.Format, new List<IFormatter>());
            }
            Formatters[formatter.Format].Add(formatter);
          } catch (MissingMethodException) {
            Console.WriteLine("WARNING: Could not instantiate {0}", t.FullName);
          }          
        }
        if (t.GetInterface(typeof (IDecomposer).FullName) != null) {
          try {
            Decomposers.Add((IDecomposer) Activator.CreateInstance(t, true));
          } catch (MissingMethodException) {
            Console.WriteLine("WARNING: Could not instantiate {0}", t.FullName);
          }
        }
      }
    }

    public Configuration GetDefaultConfig(IFormat format) {
      Dictionary<Type, IFormatter> formatterConfig = new Dictionary<Type, IFormatter>();
      foreach ( IFormatter f in Formatters[format] ) {
        if ( ! formatterConfig.ContainsKey(f.Type) )
          formatterConfig.Add(f.Type, f);
      }
      return new Configuration(formatterConfig, Decomposers);
    }

    public Configuration GetConfiguration(IFormat format) {
      if (customConfigurations.ContainsKey(format))
        return customConfigurations[format];
      return GetDefaultConfig(format);
    }

    public void DefineConfiguration(IFormat format, Configuration configuration) {
      customConfigurations[format] = configuration;
    }

  }
  
}